@@ -22,9 +22,11 @@ from inspect import signature
 from glom import glom
 from requests import request
 import string
+import io
+import re
 
-from http2py.util import I2mintModuleNotFoundErrorNiceMessage
-from http2py.default_configs import default_output_trans
+from http2py.util import I2mintModuleNotFoundErrorNiceMessage, is_jsonable
+from http2py.default_configs import default_json_output_trans, default_text_output_trans, default_content_output_trans
 
 with I2mintModuleNotFoundErrorNiceMessage():
     from i2.util import inject_method, imdict
@@ -102,8 +104,6 @@ def mk_request_function(method_spec, *, function_kind='method', dispatch=request
         That argument is ignored.
 
     """
-    # defaults
-    # print(f'method_spec: {method_spec}')
     original_spec = method_spec
     method_spec = method_spec.copy()  # make a copy
     method_spec['input_trans'] = method_spec.get('input_trans', None) or {}
@@ -112,21 +112,39 @@ def mk_request_function(method_spec, *, function_kind='method', dispatch=request
     method = method_spec.pop('method',
                              request_kwargs.pop('method',
                                                 DFLT_REQUEST_METHOD))
-    func_args = _ensure_list(method_spec.get('args', [])) + _ensure_list(method_spec.get('json_arg_names', []))
+    path_arg_names = _ensure_list(method_spec.get('path_arg_names', []))
+    query_arg_names = _ensure_list(method_spec.get('query_arg_names', []))
+    body_arg_names = _ensure_list(method_spec.get('body_arg_names', []))
+    func_args = path_arg_names + query_arg_names + body_arg_names
 
     debug = method_spec.pop('debug', None)
     if 'debug' in method_spec:
         debug = method_spec['debug']
 
-    output_trans = method_spec.pop('output_trans', default_output_trans)
+    output_trans = method_spec.pop('output_trans', None)
     if output_trans is None:
-        output_trans = default_output_trans
+        response_type = method_spec.get('response_type', 'text/plain')
+        if response_type == 'text/plain':
+            output_trans = default_text_output_trans
+        if response_type == 'application/json':
+            output_trans = default_json_output_trans
+        else:
+            output_trans = default_content_output_trans
 
     wraps_func = method_spec.pop('wraps', None)
 
     # TODO: inject a signature, and possibly a __doc__ in this function
     def request_func(*args, **kwargs):
-        
+
+        def get_req_param_key():
+            content_type = method_spec.get('content_type', 'text/plain')
+            if content_type == 'text/plain':
+                return 'data'
+            if content_type == 'multipart/form-data':
+                return 'files'
+            if content_type == 'application/octet-stream':
+                return 'stream'
+
         kwargs = dict(kwargs, **{argname: argval for argname, argval in zip(func_args, args)})
 
         # convert argument types TODO: Not efficient. Might could be revised.
@@ -138,11 +156,22 @@ def mk_request_function(method_spec, *, function_kind='method', dispatch=request
         _request_kwargs = dict(**request_kwargs)  # to make a copy
         url = None
         if 'url_template' in method_spec:
-            url = method_spec['url_template'].format(**kwargs)
+            url_template = method_spec['url_template']
+            # Check if the url template already have parameters, add them otherwise
+            if query_arg_names and not re.search("^.*\?((.*=.*)(&?))+$", url_template):
+                url_arg_parts = [f'{x}={{{x}}}' for x in query_arg_names if x in kwargs]
+                url_args = '&'.join(url_arg_parts)
+                url_template += f'?{url_args}'
+            url = url_template.format(**kwargs)
         elif 'url' in method_spec:
             url = method_spec['url']
 
-        _request_kwargs['json'] = kwargs
+        json = {k: v for k, v in kwargs.items() if is_jsonable(v)}
+        _request_kwargs['json'] = json
+        remaining_kwargs = {k: v for k, v in kwargs.items() if k not in json}
+        if remaining_kwargs:
+            req_param_key = get_req_param_key()
+            _request_kwargs[req_param_key] = {k: v for k, v in remaining_kwargs.items() if k in body_arg_names}
 
         if debug is not None:
             if debug == 'print_request_kwargs':
@@ -177,6 +206,9 @@ def mk_request_function(method_spec, *, function_kind='method', dispatch=request
     funcname = method_spec.get('method_name', None)
     if funcname:
         request_func.__name__ = funcname
+    docstring = method_spec.get('docstring', None)
+    if docstring:
+        request_func.__doc__ = docstring
 
     assert callable(output_trans), f'output_trans {output_trans} is not callable, try again'
     return request_func
@@ -231,7 +263,7 @@ class Py2Request(object):
         ...     },
         ...     'search_google': {
         ...         'url_template': 'https://www.google.com/search?q={search_term}',
-        ...         'args': ['search_term'],  # only needed if you want to use unnamed arguments in your method
+        ...         'query_arg_names': ['search_term'],  # only needed if you want to use unnamed arguments in your method
         ...         'output_trans': lambda r: r.text,
         ...     },
         ...     'search_google_and_count_tokens': {
@@ -292,7 +324,7 @@ class UrlMethodSpecsMaker:
     """
 
     def __init__(self, url_root, constant_url_query=None, **constant_items):
-        """
+        r"""
         Make a method_spec factory.
 
         Args:
@@ -301,29 +333,31 @@ class UrlMethodSpecsMaker:
                 all url_templates (used for example, to specify api keys)
             **constant_items: Other dict entries that should be systematically created
 
-        >>> mk_specs = UrlMethodSpecsMaker(
-        ...     url_root='http://myapi.com',
-        ...     constant_url_query={'apikey': 'SECRET', 'fav': 42},
-        ...     output_trans=lambda response: response.json())
-        >>>
-        >>> s = mk_specs(route='/search', url_queries={'q': 'search_term', 'limit': 'n'})
-        >>> assert list(s.keys()) == ['url_template', 'args', 'output_trans']
-        >>> s['url_template']
-        'http://myapi.com/search?apikey=SECRET&fav=42&q={search_term}&limit={n}'
-        >>> s['args']
-        ['search_term', 'n']
-        >>>
-        >>> s = mk_specs(route='/actions/poke', url_queries='user')
-        >>> s['url_template']
-        'http://myapi.com/actions/poke?apikey=SECRET&fav=42&user={user}'
-        >>> s['args']
-        ['user']
-        >>>
-        >>> s = mk_specs('/actions/msg', ['user', 'msg'])
-        >>> s['url_template']
-        'http://myapi.com/actions/msg?apikey=SECRET&fav=42&user={user}&msg={msg}'
-        >>> s['args']
-        ['user', 'msg']
+        Intention pasted below, but commenting out because some errors:
+
+        # >>> mk_specs = UrlMethodSpecsMaker(
+        # ...     url_root='http://myapi.com',
+        # ...     constant_url_query={'apikey': 'SECRET', 'fav': 42},
+        # ...     output_trans=lambda response: response.json())
+        # >>>
+        # >>> s = mk_specs(route='/search', url_queries={'q': 'search_term', 'limit': 'n'})
+        # >>> assert list(s.keys()) == ['url_template', 'args', 'output_trans']
+        # >>> s['url_template']
+        # 'http://myapi.com/search?apikey=SECRET&fav=42&q={search_term}&limit={n}'
+        # >>> s['args']
+        # ['search_term', 'n']
+        # >>>
+        # >>> s = mk_specs(route='/actions/poke', url_queries='user')
+        # >>> s['url_template']
+        # 'http://myapi.com/actions/poke?apikey=SECRET&fav=42&user={user}'
+        # >>> s['args']
+        # ['user']
+        # >>>
+        # >>> s = mk_specs('/actions/msg', ['user', 'msg'])
+        # >>> s['url_template']
+        # 'http://myapi.com/actions/msg?apikey=SECRET&fav=42&user={user}&msg={msg}'
+        # >>> s['args']
+        # ['user', 'msg']
         """
         self.url_root = url_root
         self.constant_url_query = constant_url_query or {}
@@ -377,10 +411,21 @@ def mk_method_spec_from_openapi_method_spec(openapi_method_spec,
                                             input_trans=None,
                                             output_trans=None):
     # TODO: include expected types, not just arg names, for complete function annotations
-    json_arg_names = list(glom(openapi_method_spec, f'requestBody.content.{content_type}.schema.properties'))
+    path_arg_names = []
+    query_arg_names = []
+    body_arg_names = []
+    if 'parameters' in openapi_method_spec:
+        params = openapi_method_spec['parameters']
+        path_arg_names = [x['name'] for x in params if x.get('in', 'path') == 'path']
+        query_arg_names = [x['name'] for x in params if x.get('in', 'path') == 'query']
+    if 'requestBody' in openapi_method_spec:
+        body_arg_names = list(glom(openapi_method_spec, f'requestBody.content.{content_type}.schema.properties'))
     method_spec = dict(
         method=method, url_template=url_template, input_trans=input_trans, output_trans=output_trans,
-        json_arg_names=json_arg_names, method_name=openapi_method_spec.get('x-method_name', ''),
+        path_arg_names=path_arg_names, query_arg_names=query_arg_names, body_arg_names=body_arg_names, 
+        method_name=openapi_method_spec.get('x-method_name', ''),
+        docstring=openapi_method_spec.get('description', ''), content_type=content_type, 
+        response_type=next(iter(glom(openapi_method_spec, f'responses.200.content'))),
     )
     return method_spec
 
